#' Bias Correction based on the mean and standard deviation adjustment following Torralba et al. (2017) 
#'
#'@author Verónica Torralba, \email{veronica.torralba@bsc.es}
#'@description This function applies the simple bias adjustment technique described in Torralba et al. (2017). The adjusted forecasts have an equivalent standard deviation and mean to that of the reference dataset.
#'#'@param data a list of s2dverification objects (lists) as output by the \code{Load} function from the s2dverification package.
#'
#'@return \code{$biasCorrected} {An array with the bias corrected forecasts with the dimensions of data$mod.}
#'
#'@import s2dverification
#'@examples
#'
#'# Example
#'# Creation of sample s2dverification objects. These are not complete
#'# s2dverification objects though. The Load function returns complete objects.
#'mod1 <- 1 : (1 * 3 * 4 * 5 * 6 * 7)
#'dim(mod1) <- c(dataset = 1, member = 3, sdate = 4, ftime = 5, lat = 6, lon = 7)
#'obs1 <- 1 : (1 * 1 * 4 * 5 * 6 * 7)
#'dim(obs1) <- c(dataset = 1, member = 1, sdate = 4, ftime = 5, lat = 6, lon = 7)
#'lon <- seq(0, 30, 5)
#'lat <- seq(0, 25, 5)
#'data1 <- list(mod = mod1, obs = obs1, lat = lat, lon = lon)
#'a <- BiasCorrection(data1)
#'str(a)
#'@export


BiasCorrection <- function(data) {
  is_s2dv_object <- function(x) {
    if (all(c('mod', 'obs', 'lon', 'lat') %in% names(x))) { #&& length(x) == 11) {
      TRUE
    } else {
      FALSE
    }
  }
  wrong_input <- FALSE
  if (!is.list(data)) {
    wrong_input <- TRUE
  }
  
  if (!is_s2dv_object(data)) {
    wrong_input <- TRUE
  }

  if (length(dim(data$obs)) != 6) {
    stop('data$obs must have six dimensions')
  }
  
  if (length(dim(data$mod)) != 6) {
    stop('data$mod must have six dimensions')
  }

  if (wrong_input) {
    stop("Parameter 'data' must be a list of s2dverification objects as returned by ",
         "the s2dverification::Load function.")
  }
  
  if (length(which(is.na(data$mod))) > 0) {
    warning('There are NA in data$mod')
  }
  
  if (length(which(is.na(data$obs))) > 0) {
    warning('There are NA in data$obs')
  }
  
  biasAdjusted <- Apply(data = list(var_obs = data$obs, var_exp = data$mod),
                      target_dims = list(c('member', 'sdate'), c('member', 'sdate')),
                      fun = '.sbc')
  BiasCorrected <- aperm(biasAdjusted$output1, c(3, 1, 2, 4, 5, 6))
  names(dim(BiasCorrected)) <- names(dim(data$mod))
  return(list(biasCorrected = BiasCorrected))
}

.sbc <- function(var_obs, var_exp) {
  ntime <- length(var_obs)
    if (dim(var_exp)[1]!=dim(var_exp)[2]){
    nmembers <- dim(var_exp)[-which(dim(var_exp) == length(var_obs))]
  }else{
    nmembers<-dim(var_exp)[1]
  }
  
  if (!all(dim(var_exp)== c(nmembers, ntime))) {
    var_exp <- t(var_exp)
  }
  
  corrected <- NA * var_exp
  
  for (t in 1:ntime) {
    # defining forecast,hindcast and observation in cross-validation
    fcst <- var_exp[, t]
    hcst <- var_exp[,-t]
    obs <- var_obs[-t]
    
    # parameters
    sd_obs <- sd(obs)
    sd_exp <- sd(hcst)
    clim_exp <- mean(hcst)
    clim_obs <- mean(obs)
    
    # bias corrected forecast
    corrected[, t] <- ((fcst - clim_exp) * (sd_obs / sd_exp)) + clim_obs
  }
  
  return(corrected)
}
